import argparse as ap
import logging
import logging.handlers
import os
import shutil
from pathlib import Path
from tarfile import TarFile

import cx_Oracle as almadb
import requests
import sys
from ._version import ___version___ as version
from pymygdala import LogHandler
from pycapo import CapoConfig

"""
    A utility script for accessing ALMA calibration products, based largely upon the JAO listfiles.py script.

    It is assumed that this is running in the products subdirectory of a 'typical' CASA processing layout. 

    1) Query the metadata & ngas databases at the NAASC
    2) Retrieve the desired files, verifying against the NGAS size  
    3) Arrange the files for a restore
        NOTE: assumes standard CASA layout (i.e. ../rawdata exists)
        a) unzip the auxproducts file 
        b) copy the pipeline_manifest file to the rawdata directory (../rawdata)
"""

# Define our query:  Notice that it crosses 2 databases on the same server
_CAL_PRODUCTS_QUERY = """select FILE_ID,FILE_SIZE from ngas.ngas_files where file_id in 
                            (select ngas_file_id from alma.asa_product_files where ASA_OUS_ID = :mous_uid  
                            and FILE_CLASS in ('calibration','script'))"""

# Tool Description for --help
_DESCRIPTION = """AAT/PPI tool for acquiring & arranging ALMA calibration products to perform a restore. Version {}"""

# set up logging:
logger = logging.getLogger("alma_product_fetcher")
logger.setLevel(logging.DEBUG)

config = CapoConfig()


def _make_parser():
    r""" Build a command line parser for this app: this is external to MrClean and/or CapoSettings because both need
    it, and Sphinx will want a look at ot to build docs. """

    result = ap.ArgumentParser(description=_DESCRIPTION.format(version),
                               formatter_class=ap.RawTextHelpFormatter)
    result.add_argument('-P', '--profile', action='store', default='',
                        help='profile name to use, e.g. test, production')
    result.add_argument('-n', '--noArrangement', action='store_true',
                        help='Just download the files, do not perform the extra arrangement step')
    # result.add_argument('-e', '--email', type=str, action='store', default='', help='Email for completion notification')
    result.add_argument('mousUid', type=str, action='store',
                        help='name of the MOUS for which to acquire products. (ex. uid://A003/Xaaaa/Xbbbb)')
    return result


# Connects to the ALMA database
# Blatantly stolen from alma_reingester
def connect_alma():
    user = config['almaMetadataDatabase.jdbcUsername']
    passwd = config['almaMetadataDatabase.jdbcPassword']
    url = config['almaMetadataDatabase.jdbcUrl']

    hostport, service_name = url.split('/')[-2:]
    host, port = hostport.split(':')
    dsn = almadb.makedsn(host, port, service_name=service_name.upper())
    return almadb.connect(user, passwd, dsn)


# Slightly modified from the original source: formatted table printing
# of name & size
def justified_print(list):
    if len(list) > 0:
        maxlen0 = max([len(item[0]) for item in list])
        maxlen1 = max([len(str(item[1])) for item in list])

        logger.debug('\n')
        logger.debug('{0:s}\t{1:s}'.format(str.ljust('FILE_ID', maxlen0), str.rjust('FILE_SIZE', maxlen1)))

        for i in list:
            logger.debug('{0:s}\t{1:s}'.format(str.ljust(i[0], maxlen0), str.rjust(str(i[1]), maxlen1)))
    else:
        logger.warning('No files found')


# Example from MOUS uid://A001/X1284/X265f:
#
# FILE_ID                                                              	FILE_SIZE
# uid___A002_Xc7111c_X86a7.ms.calapply.txt                             	   3864
# uid___A002_Xc7111c_X86a7.ms.flagversions.tgz                         	 128436
# uid___A002_Xc7111c_X86a7_target.ms.auxcalapply.txt                   	    623
# uid___A002_Xc72427_X2dfe.ms.calapply.txt                             	   3867
# uid___A002_Xc72427_X2dfe.ms.flagversions.tgz                         	 128729
# uid___A002_Xc72427_X2dfe_target.ms.auxcalapply.txt                   	    623
# member.uid___A001_X1284_X265f.scriptForPI.py                         	  17756
# member.uid___A001_X1284_X265f.calimage.pipeline_manifest.xml         	  11057
# member.uid___A001_X1284_X265f.hifa_calimage.auxproducts.tgz          	   1851
# member.uid___A001_X1284_X265f.hifa_calimage.casa_piperestorescript.py	    354
# member.uid___A001_X1284_X265f.hifa_calimage.casa_pipescript.py       	   3389
# member.uid___A001_X1284_X265f.session_1.auxcaltables.tgz             	2057848
# member.uid___A001_X1284_X265f.session_1.caltables.tgz                	4587123
# member.uid___A001_X1284_X265f.session_2.auxcaltables.tgz             	2590616
# member.uid___A001_X1284_X265f.session_2.caltables.tgz                	6753280
# member.uid___A001_X1284_X265f.session_3.auxcaltables.tgz             	2094142
# member.uid___A001_X1284_X265f.session_3.caltables.tgz                	4598380
# member.uid___A001_X1284_X265f.session_4.auxcaltables.tgz             	2715062
# member.uid___A001_X1284_X265f.session_4.caltables.tgz                	4583153
# uid___A002_Xc7111c_X3979.ms.calapply.txt                             	   3862
# uid___A002_Xc7111c_X3979.ms.flagversions.tgz                         	 128920
# uid___A002_Xc7111c_X3979_target.ms.auxcalapply.txt                   	    623
# uid___A002_Xc7111c_X409e.ms.calapply.txt                             	   3865
# uid___A002_Xc7111c_X409e.ms.flagversions.tgz                         	 158358
# uid___A002_Xc7111c_X409e_target.ms.auxcalapply.txt                   	    624


def lookup_products_files(mous_id):
    r"""Perform the multi-stage query required for the list of desired files, and return that list
    for further processing."""

    alma_handle = connect_alma().cursor()

    alma_handle.execute(_CAL_PRODUCTS_QUERY, mous_uid=mous_id)

    results = alma_handle.fetchall()

    return results


def retrieve_from_ngas(filename, expected_size):
    r"""Simple request to the NAASC NGAS master node.  Take a look at the responses."""

    ngas_server = config['almaNgasSystem.serverName']  # 'nangas08'
    ngas_port = config['almaNgasSystem.serverPort']  # '7773'

    ngas_url = 'http://{}.cv.nrao.edu:{}/RETRIEVE?file_id={}'.format(ngas_server, ngas_port, filename)

    # logger.info("NGAS Request: {}".format(ngas_url))

    answer = requests.get(ngas_url)

    if requests.codes.ok == answer.status_code:

        header_ratio = int(answer.headers['Content-length']) / expected_size

        # if the header doesn't return the right size, don't write the file.
        if header_ratio != 1.0:
            logger.error("E: {}\tH: {}".format(expected_size, answer.headers['Content-length']))
            logger.error("Expected equality of these values, not streaming the file.")
            return False

        # Write out the filename, track what's written
        accumulator = 0
        with open(filename, 'wb') as fd:
            for chunk in answer.iter_content(chunk_size=128):
                # write returns the number of bytes
                accumulator += fd.write(chunk)

        written_ratio = expected_size / accumulator

        # logger.info(
        #    "Expected: {}\tWritten:{}\tRatio: {}".format(expected_size, accumulator, expected_size / accumulator))

        return written_ratio == 1.0

    else:
        logger.error('Failed request!')
        return False


def obtain_calibration_products(mous_id):

    file_list = lookup_products_files(mous_id)

    for [filename, size] in file_list:
        ok = retrieve_from_ngas(filename, size)

        if not ok:
            logger.error('Retrieval of {} failed!'.format(filename))
            # What should we do here?  currently just continuing on....

    logger.info('Retrieved all the required items for {}.'.format(mous_id))


def perform_final_restore_setup():
    r"""Final details:  Copy the pipeline manifest to ../rawdata, and
    unzip the *.auxproducts.tgz file to allow the pipeline to access
    the CSV files it needs.

    NOTE: With the split of calibration & imaging into separate steps, we need to be more explict about the
    names (require 'hifa_cal' but allow for 'hifa_calimage' as well

    """

    # Get our current location.
    cwd = Path.cwd()

    # And where the manifest file needs to go:
    destination = cwd / ".." / "rawdata"
    destination.resolve()

    if not destination.exists():
        logger.warning("Need to create the path {}.".format(str(destination)))
        destination.mkdir()

    #
    # Now find & copy over the manifest
    #
    manifest_file = list(cwd.glob('*_cal*.pipeline_manifest.xml'))

    # There should be only one manifest, otherwise things go bad.
    if len(manifest_file) != 1:
        logger.error("More than one Manifest file, the pipeline won't run.")
        return

    #
    source_file = cwd / manifest_file[0]
    dest_file = destination / manifest_file[0]
    if not dest_file.exists():
     shutil.copy(str(source_file), str(dest_file))

    #
    # Now we need to unzip the .auxproducts.tgz file in place:
    #

    aux_products = list(cwd.glob('*_cal*.auxproducts.tgz'))

    if len(aux_products) != 1:
        logger.error("Multiple auxproducts files.  That shouldn't happen!.")
        return

    # Open, list the contents for completeness sake, extract in place.
    tar_handle = TarFile.open(name=aux_products[0])
    tar_handle.list()
    tar_handle.extractall()


def main():

    # lookup_products_files('uid://A001/X1284/X265f')
    # retrieve_from_ngas('uid___A002_Xc7111c_X3979.ms.calapply.txt', 3862)
    # obtain_calibration_products('uid://A001/X1284/X265f')

    # Get our arguments
    args = _make_parser().parse_args()

    # Shamelessly stolen from epilogue with a twist: allow for explict profile setting via the CL
    if 'CAPO_PROFILE' not in os.environ and '' == args.profile:
        # try to synthesize a profile from our installation root
        profile = os.path.abspath(sys.argv[0]).split(os.path.sep)[-3]
        os.environ['CAPO_PROFILE'] = profile
        print('No CAPO_PROFILE explicitly set, synthesizing {0} by path'.format(str(profile)))
    elif '' != args.profile:
        os.environ['CAPO_PROFILE'] = args.profile

    # integrate the logs with pymygdala's logging system
    handler = LogHandler(profile=args.profile, application='alma_product_fetcher')
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)

    files = lookup_products_files(args.mousUid)

    logger.info("Found {} calibration products & scripts for {}".format(len(files), args.mousUid))

    # justified_print(files)

    obtain_calibration_products(args.mousUid)

    if args.noArrangement:
        logger.debug("No further work to be done.")
    else:
        perform_final_restore_setup()
        logger.debug("Completed restore setup.")
