import logging

from sqlalchemy import text
from sqlalchemy.orm import aliased

from pyat.schema import create_session
from pyat.schema.model import Author, Project, ExecutionBlock, Filegroup, File, \
    Polarization, Configuration, DataDescription, AlmaOus, AlmaOusType, Calibration, \
    ScienceProduct, ScienceProductsProductGroup, ProductGroup

session = create_session('SDM')
LOG = logging.getLogger('ingestion')


def getpolarizationcode(polarization):
    polarizationid = session.query(Polarization.polarization_id).\
        filter(text("name=:polarization")).\
        params(polarization=polarization).\
        one_or_none()
    if polarizationid is None:
        return 0
    session.close()
    return polarizationid[0]


def getproject(projectcode):
    project = session.query(Project).\
        filter(Project.project_code==projectcode).\
        one_or_none()
    if project is not None:
        project.authors
    session.close()
    return project


def getauthor(projectcode, personid):
    author = session.query(Author). \
        filter(Author.project_code==projectcode). \
        filter(Author.pst_person_id==str(personid)). \
        first()
    session.close()
    return author


def get_filegroup_from_name(external_name):
    science_product = session.query(ScienceProduct). \
        filter(ScienceProduct.external_name==external_name). \
        one_or_none()

    return science_product.filegroup


def get_file_from_group(group):
    file = session.query(File). \
        filter(File.filegroup1 == group). \
        one_or_none()
    return file


def getexecutionblockfromname(ngas_name):
    executionblock = session.query(ExecutionBlock). \
        filter(ExecutionBlock.ngas_fileset_id.isnot(None)). \
        filter(ExecutionBlock.ngas_fileset_id==ngas_name). \
        one_or_none()
    return executionblock


def get_execution_block_from_external_name(external_name):
    """
    Get execution block with the given external name
    :param external_name: external name of the execution block
    :return: execution block
    """
    execution_block = session.query(ExecutionBlock). \
        filter(ScienceProduct.external_name==external_name). \
        filter(ExecutionBlock.science_product_locator==ScienceProduct.science_product_locator). \
        one_or_none()

    return execution_block


def get_execution_block(science_product_locator):
    execution_block = session.query(ExecutionBlock). \
        filter(ExecutionBlock.ngas_fileset_id.isnot(None)). \
        filter(ExecutionBlock.science_product_locator==science_product_locator). \
        one_or_none()
    session.close()
    return execution_block


def get_science_product_by_spl(science_product_locator):
    science_product = session.query(ScienceProduct). \
        filter(ScienceProduct.science_product_locator==science_product_locator). \
        one_or_none()
    return science_product


def get_science_product(filename):
    """
    Get the science product associated with this file
    :param filename:
    :return:
    """
    science_product = session.query(ScienceProduct). \
        filter(File.filename==filename). \
        filter(File.filegroup==Filegroup.filegroup_id). \
        filter(Calibration.filegroup_id==Filegroup.filegroup_id). \
        filter(ScienceProduct.science_product_locator==Calibration.science_product_locator). \
        one_or_none()
    return science_product


def get_file(filename):
    file = session.query(File). \
        filter(File.filename==filename). \
        one_or_none()
    return file


def getvlbaexecutionblock(project_code, segment):
    executionblock = session.query(ExecutionBlock). \
        filter(ExecutionBlock.project_code==project_code). \
        filter(ExecutionBlock.segment==segment). \
        one_or_none()
    session.close()
    return executionblock


def get_vlba_correlations_filegroup(project_code, segment, filegroup_type):
    filegroup = session.query(ExecutionBlock, Filegroup). \
        filter(ExecutionBlock.project_code == project_code). \
        filter(ExecutionBlock.segment == segment). \
        filter(ExecutionBlock.filegroup_id == Filegroup.filegroup_id). \
        filter(Filegroup.type == filegroup_type). \
        one_or_none()
    if filegroup is None:
        LOG.debug(f"Filegroup type {filegroup_type} not found for segment {project_code}{segment}.")
    session.close()
    return filegroup


def get_vlba_correlator_pass_filegroup(project_code, segment, filegroup_type):
    eb, filegroup = session.query(ExecutionBlock, Filegroup). \
        filter(ExecutionBlock.project_code == project_code). \
        filter(ExecutionBlock.segment == segment). \
        filter(ExecutionBlock.filegroup_id == Filegroup.parent_filegroup_id). \
        filter(Filegroup.type == filegroup_type). \
        one_or_none()
    if filegroup is None:
        LOG.debug(f"Filegroup type {filegroup_type} not found for segment {project_code}{segment}.")
    session.close()
    return filegroup


def getvlbafile(ngas_id):
    file = session.query(File). \
        filter(File.ngas_id == ngas_id). \
        one_or_none()
    return file


def removeexecutionblock(executionblock):
    session.delete(executionblock)
    session.commit()


def removefilegroup(filegroup):
    session.delete(filegroup)
    session.commit()


def removefile(file):
    session.delete(file)
    session.commit()


def getpolarization(polarizationcode):
    polarization = session.query(Polarization).\
        filter(Polarization.polarization_id==polarizationcode).\
        one()
    session.close()
    return polarization


def getdatadescription(bandwidth, frequency, polarization):
    datadescription = asession.query(DataDescription).\
        filter(DataDescription.bandwidth==bandwidth).\
        filter(DataDescription.frequency==frequency).\
        filter(DataDescription.polarization_id==polarization).\
        first()
    return datadescription


def getconfiguration(execblocid, configurationid, asession):
    configuration = asession.query(Configuration).\
        filter(Configuration.execution_block_id==execblocid).\
        filter(Configuration.configuration==configurationid).\
        one_or_none()
    return configuration


def getimagesetfilegroupfromname(groupname):
    filegroup = session.query(Filegroup). \
        filter(Filegroup.groupname == groupname). \
        filter(Filegroup.type == 'image_set'). \
        one_or_none()
    session.close()
    return filegroup


def removeimageset(filegroup):
    session.delete(filegroup)
    session.commit()


# Check for the existence of OUSes respectively since they will have
# to be built if they don't exist

def ousexists(id, type):
    """

    :param id: alma_ous_id
    :param type: MOUS, GOUS, or SOUS
    :return True if OUS exists in the archive db:
    """

    cnt = session.query(AlmaOus, AlmaOusType). \
        filter(AlmaOus.ous_type==AlmaOusType.ous_type). \
        filter(AlmaOus.alma_ous_id==id). \
        filter(AlmaOusType.ous_type==type). \
        count()
    return True if cnt > 0 else False


def getprojectcodefrommous(mous):
    """
    Get the project code from the MOUS using OUS hierarchy
    :param mous:
    :return: project code
    """

    LOG.info(f"Find project code for MOUS: {mous}")

    almaouss = aliased(AlmaOus)
    almaousg = aliased(AlmaOus)
    almaousm = aliased(AlmaOus)

    projectcode = session.query(almaouss.project_code). \
        join(almaousg, almaousg.parent_ous_id==almaouss.alma_ous_id). \
        join(almaousm, almaousm.parent_ous_id==almaousg.alma_ous_id). \
        filter(almaousm.alma_ous_id==mous). \
        filter(almaouss.ous_type=='SOUS'). \
        filter(almaousg.ous_type=='GOUS'). \
        filter(almaousm.ous_type=='MOUS'). \
        one_or_none()

    session.close()

    return projectcode[0]


def getmousesfromprojectcode(project_code):
    """
    Get MOUSes for a project
    :param project_code: 
    :return: list of MOUSes
    """

    mset = set()

    project = session.query(Project). \
        filter(Project.project_code==project_code). \
        one_or_none()

    for sous in project.alma_ouses:
        gouses = sous.children_ouses
        for gous in gouses:
            mouses = gous.children_ouses
            for mous in  mouses:
                mset.add(mous)

    sessopm/c;pse
    return list(mset)

def getcalibrationsfrommous(mousid):
    """
    Get calibrations for MOUS
    :param MOUS id: 
    :return list of calibrations: 
    """

    mous = session.query(AlmaOus). \
        filter(AlmaOus.alma_ous_id==mousid). \
        filter(AlmaOus.ous_type=='MOUS'). \
        one_or_none()

    session.close()

    return mous.calibrations

def getcalibrationsfromeb(ngas_fileset_id):
    """
    Get calibrations for execution block
    :param MGAS fileset id:
    :return list of calibrations:
    """

    eb = session.query(ExecutionBlock). \
        filter(ExecutionBlock.ngas_fileset_id==ngas_fileset_id). \
        one_or_none()

    session.close()
    return eb.calibrations


def getexecutionblocksfrommous(mousid):
    """
    Get execution blocks for MOUS
    :param MOUS id:
    :return list of execution_blocks:
    """

    mous = session.query(AlmaOus). \
        filter(AlmaOus.alma_ous_id==mousid). \
        filter(AlmaOus.ous_type=='MOUS'). \
        one_or_none()

    session.close()
    return mous.execution_blocks


def get_mous_from_eb_science_product_locator_list(locators):
    """
    Get the mous from a list of ebs (should be identical)
    :param locators:
    :return:
    """

    mous_set = set()
    for locator in locators:
        eb = session.query(ExecutionBlock). \
            filter(ExecutionBlock.science_product_locator==locator). \
            one_or_none()
        if eb.alma_ous is not None:
            mous_set.add(eb.alma_ous)
    if len(mous_set) > 1:
        LOG.critical("List of EB science product locators resulted in multiple MOUS")

    LOG.debug(f"MOUS set: {mous_set}")
    session.close()
    return mous_set.pop()


def get_project_from_science_product_locator(science_product_locator):
    """
    Return the project given the science product
    :param science_product_locator: Science product locator
    :return: project
    """

    science_product = session.query(ScienceProduct). \
        filter(ScienceProduct.science_product_locator==science_product_locator). \
        one_or_none()

    # session.close()
    return science_product.project
